"""
fritz.api_resources.app
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
import logging

import fritz
import fritz.utils
import fritz.errors
from fritz.api_resources.fritz_object import FritzObject

_logger = logging.getLogger(__name__)


class App(FritzObject):
    """Resources to manipulate app resources from the CLI."""

    OBJECT_NAME = "Application"

    def summary(self):
        """Print summary of object."""
        details = [
            ("Created At", self.created_date),
            ("APK ID", self.apk_id),
            ("Platform", self.platform),
            ("API Key", self.api_key),
        ]
        return self._summary(self.app_name, details)

    # pylint: disable=duplicate-code
    @classmethod
    def create(
        cls,
        api_key=None,
        project_id=None,
        app_name=None,
        apk_id=None,
        platform=None,
    ):
        """Create a new app.

        Args:
            api_key (str): Fritz API Key
            app_name (str): App Name
            apk_id (str): APK ID
            platform (str): Platform app is created on (ios or android)

        Raises:
            fritz.errors.FritzNotInitializedError

        Returns: App object
        """
        url = "/app"
        params = {
            "project_uid": project_id or fritz.project_id,
            "app_name": app_name,
            "apk_id": apk_id,
            "platform": platform,
        }
        response = cls.client_post(url, api_key=api_key, params=params)
        obj = fritz.utils.convert_to_fritz_object(response)
        return obj

    @classmethod
    def get(cls, api_key=None, app_id=None):
        """Get app.

        Args:
            api_key: Optional API Key to use.
            app_id: ID of app to query.

        Returns: App
        """
        url = "/app/" + app_id
        response = cls.client_get(url, api_key=api_key)
        return fritz.utils.convert_to_fritz_object(response)

    @classmethod
    def list(
        cls,
        api_key=None,
        project_id=None,
        platform=None,
        app_api_key=None,
        apk_id=None,
        all_projects=False,
    ):
        """Get app, restricting results to specified arguments.

        Args:
            api_key: Optional API Key to use.
            project_id: Project ID
            platform: App platform  - either ios or android.
            app_api_key: API Key for app.
            apk_id: APK ID or Bundle ID of apps.
            all_projects: If True and no project_id specified, returns apps
                from all projects. Default False.

        Returns: List[App]
        """
        url = "/app"
        if not project_id and not all_projects:
            project_id = fritz.project_id

        params = {
            "project_uid": project_id,
            "platform": platform,
            "api_key": app_api_key,
            "apk_id": apk_id,
        }
        response = cls.client_get(url, api_key=api_key, params=params)

        return [
            fritz.utils.convert_to_fritz_object(app)
            for app in response["apps"]
        ]



# pylint: disable=abstract-method
"""
cli.app.android.android_snippets

Android code snippets.
"""

from cli.app import code_snippet
from cli.app import file_manager
from cli.app.android import android_helpers


class AndroidManifestSnippet(code_snippet.CodeSnippet):
    """Base class for snippets added to AndroidManifest.xml"""

    def __init__(self, app_config, app):
        super().__init__(app_config, app)
        self.path = app_config.android_manifest_path


class MainActivitySnippet(code_snippet.CodeSnippet):
    """Base class for snippets added to the Main Activity."""

    def __init__(self, app_config, app):
        super().__init__(app_config, app)
        self.path = app_config.main_activity_path


class AppGradleSnippet(code_snippet.CodeSnippet):
    """Base class for snippets added to App build.gradle file."""

    def __init__(self, app_config, app):
        super().__init__(app_config, app)
        self.path = app_config.app_gradle_path


class ProjectGradleSnippet(code_snippet.CodeSnippet):
    """Base class for snippets added to Project build.gradle file."""

    def __init__(self, app_config, app):
        super().__init__(app_config, app)
        self.path = app_config.project_gradle_path


class FritzCustomModelServiceSnippet(AndroidManifestSnippet):
    """Adds FritzCustomModelService. """

    def __init__(self, app_config, app):
        super().__init__(app_config, app)
        self.path = app_config.android_manifest_path

    @property
    def is_added(self):
        contents = self.path.read_text()
        return "ai.fritz.core.FritzCustomModelService" in contents

    def add(self):
        snippet = """
        <service
            android:name="ai.fritz.core.FritzCustomModelService"
            android:exported="true"
            android:permission="android.permission.BIND_JOB_SERVICE" />\n
        """
        modified = file_manager.insert_before(
            self.path, "</application>", snippet
        )
        self.path.write_text(modified)


class InternetPermissionsSnippet(AndroidManifestSnippet):
    """Adds internet permissions to manifest."""

    @property
    def is_added(self):
        contents = self.path.read_text()
        return "android.permission.INTERNET" in contents

    def add(self):
        snippet = """
    <uses-permission android:name="android.permission.INTERNET" />
        """
        modified = file_manager.insert_before(
            self.path, "<application", snippet
        )
        self.path.write_text(modified)


class ImportFritzSnippet(MainActivitySnippet):
    """Import Fritz in main activity."""

    @property
    def is_added(self):
        contents = self.path.read_text()
        return "import ai.fritz.core.Fritz;" in contents

    def add(self):
        modified = file_manager.insert_after(
            self.path, "package", "import ai.fritz.core.Fritz;\n"
        )
        self.path.write_text(modified)


class ConfigureFritzSnippet(MainActivitySnippet):
    """Adds Fritz.configure call in Main Activity."""

    @property
    def is_added(self):
        contents = self.path.read_text()
        return "Fritz.configure" in contents

    def add(self):
        configure = file_manager.add_indentation(
            2
        ) + 'Fritz.configure(this, "{api_key}");'.format(
            api_key=self.app.api_key
        )
        modified = file_manager.insert_line_at_block_end(
            self.path, "void onCreate", configure
        )
        self.path.write_text(modified)

    @property
    def is_updated(self):
        if not self.is_added:
            return True

        api_key = android_helpers.find_fritz_api_key(self.path)
        return api_key == self.app.api_key

    def update(self):
        api_key = android_helpers.find_fritz_api_key(self.path)
        text = self.path.read_text()
        updated = text.replace(api_key, self.app.api_key)
        self.path.write_text(updated)


class AddBuildDependencies(AppGradleSnippet):
    """Add Fritz build dependendies to App Gradle file."""

    @property
    def is_added(self):
        contents = self.path.read_text()
        return "ai.fritz:core" in contents

    def add(self):
        text = """
    implementation 'ai.fritz:core:+'
    implementation 'ai.fritz:vision:+'
        """
        modified = file_manager.insert_line_at_block_end(
            self.path, "dependencies", text
        )
        self.path.write_text(modified)


class AddRenderscript(AppGradleSnippet):
    """Enable renderscript in the app Gradle file."""

    @property
    def is_added(self):
        contents = self.path.read_text()
        return (
            "renderscriptTargetApi" in contents
            and "renderscriptSupportModeEnabled" in contents
        )

    def add(self):
        text = """
        renderscriptTargetApi 21
        renderscriptSupportModeEnabled true
        """
        modified = file_manager.insert_line_at_block_end(
            self.path, "defaultConfig", text
        )
        self.path.write_text(modified)


class AddProjectBuildDependencies(ProjectGradleSnippet):
    """Add Fritz repository to Project Gradle file."""

    @property
    def is_added(self):
        contents = self.path.read_text()
        repo = "https://raw.github.com/fritzlabs/fritz-repository/master"
        return repo in contents

    def add(self):
        lines = self.path.open().readlines()
        start, end = file_manager.find_block_indexes(lines, "allprojects")

        if start is None or end is None:
            msg = """
Cannot find allprojects in project build.gradle.

Please add allprojects and run setup again.
"""
            raise code_snippet.SnippetExecutionError(msg)

        repo_start, _ = file_manager.find_block_indexes(
            lines[start : end + 1], "repositories"
        )
        if repo_start is None:
            msg = """
Cannot find repositories in project build.gradle.

Please add a repositories section in allproject.
"""
            raise code_snippet.SnippetExecutionError(msg)

        code = """        maven {
            url 'https://raw.github.com/fritzlabs/fritz-repository/master'
        }
"""
        lines.insert(start + repo_start + 1, code)
        self.path.write_text("".join(lines))



"""
fritz.api_resources.model_version
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
import os
import time
import logging
import json
import math
import requests

import fritz
import fritz.utils
import fritz.errors
from fritz.api_resources.fritz_object import FritzObject
from fritz.api_client import FritzClient

_logger = logging.getLogger(__name__)


def _convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    to_raise = int(math.floor(math.log(size_bytes, 1024)))
    power = math.pow(1024, to_raise)
    size = round(size_bytes / power, 2)
    return "%s %s" % (size, size_name[to_raise])


class ModelVersion(FritzObject):
    """ModelVersion  """

    OBJECT_NAME = "ModelVersion"

    def summary(self):
        """Print summary of object."""
        details = [
            ("Model Size", self.model_size),
            ("Filename", self.provided_filename),
            ("Version", self.version_number),
            ("Metadata", json.dumps(self.metadata, indent=2)),
            ("Loss", self.loss),
        ]
        return self._summary(self.provided_filename, details)

    @property
    def model_size(self):
        """Convert model bytes to human readable string."""
        return _convert_size(self.model_bytes)

    @property
    def loss(self):
        """Training loss if in metadata."""
        metadata = self.metadata or {}
        return metadata.get("loss")

    # pylint: disable=duplicate-code
    @classmethod
    def create(
        cls,
        api_key=None,
        project_id=None,
        model_id=None,
        snapshot_id=None,
        filename=None,
        data=None,
        set_active=None,
        metadata=None,
    ):
        """Create a new ModelVersion, uploading file to Fritz.

        Args:
            api_key (Optional[str]): API Key. Optional if `fritz.configure`
                was called.
            project_id (Optional[str]): Project ID. Optional if
                `fritz.configure` was called with project_id.
            model_id (str): optional Model ID.
            snapshot_id (str): If snapshot_id set, will attach version to
                given snapshot.
            filename (str): Name of model version filename.
            data (io.BytesIO): Model data to upload
            set_active (bool): If True, model will be set as the active version
               in Fritz. If it is True, any devices runninng this model will
               download the latest version.
            metadata (dict): Dictionary of metadata.

        Returns:
            Tuple[ModelVersion, ModelSnapshot, Model]
        """
        if not fritz.api_key and not api_key:
            raise fritz.errors.FritzNotInitializedError()

        if model_id:
            url = "/model/{model_id}/version".format(model_id=model_id)
        else:
            url = "/model/version"

        client = FritzClient(api_key or fritz.api_key)
        response = client.post(
            url,
            params={
                "project_uid": project_id or fritz.project_id,
                "snapshot_uid": snapshot_id,
                "model_uid": model_id,
            },
            data={
                "set_active": set_active,
                "metadata_json": json.dumps(metadata or {}),
            },
            files={"file": (filename, data)},
        )
        converted = {
            key: fritz.utils.convert_to_fritz_object(value)
            for key, value in response.items()
        }
        return converted["version"], converted["snapshot"], converted["model"]

    @classmethod
    def get(cls, version_id=None, api_key=None):
        """Get version by version_id

        Args:
            version_id: Version ID
            api_key (str): Optional API Key to use.

        Returns: ModelVersion
        """
        if not version_id:
            raise fritz.errors.ArgumentRequiredError("version_id")

        url = "/model/version/{version_id}".format(version_id=version_id)

        response = cls.client_get(url, api_key=api_key)
        return fritz.utils.convert_to_fritz_object(response)

    @classmethod
    def list(cls, model_id, api_key=None):
        """Get all model versions for model_id

        Args:
            model_id: Model ID
            api_key (str): Optional API Key to use.

        Returns: List[ModelVersion]
        """
        url = "/model/" + model_id + "/version"
        response = cls.client_get(url, api_key=api_key)
        return [
            fritz.utils.convert_to_fritz_object(version)
            for version in response["versions"]
        ]

    def benchmark(self, api_key=None, wait_seconds=5, attempts=5):
        """Get model grade report, waiting if it does not yet exist.

        Args:
            api_key (str): Optional API Key
            wait_seconds (int): Number of seconds to wait between each
                request.
            attempts (int): Number of attempts to make.

        Returns: ModelGradeReport if it exists
        """

        try:
            return fritz.ModelGradeReport.get(
                api_key=api_key, version_id=self.id
            )
        except fritz.errors.FritzError as err:
            if err.status_code != 404:
                raise err
            _logger.info("Grader report not found, trying again")
            attempts -= 1
            if attempts < 0:
                raise err

            time.sleep(wait_seconds)

            return self.benchmark(
                api_key=api_key, wait_seconds=wait_seconds, attempts=attempts
            )

    def download(self, download_dir=None):
        """Download Model version.

        Args:
            download_dir (str): optional directory to download model in.

        Returns: str path of downloaded model.
        """
        response = requests.get(self.s3_path, allow_redirects=True)
        path = os.path.join(download_dir or ".", self.provided_filename)
        with open(path, "wb") as f:
            f.write(response.content)
        return path



"""
cli.app.ios.ios_helpers
"""

from typing import Optional
import plistlib
import pathlib
import os
from pbxproj import XcodeProject
import fritz
from cli import print_utils


def read_plist(plist_path):
    """Read plist file.

    Args:
        plist_path (str): Path to plist.

    Returns: Dict if exists and is readable, None if error encountered.
    """
    try:
        # TODO: Replace with non-deprecated method.
        # pylint: disable=deprecated-method
        plist = plistlib.readPlist(str(plist_path))
        return plist
    # pylint: disable=broad-except
    except (plistlib.InvalidFileException, Exception):
        print_utils.warn(
            "{plist_path} does not exist".format(plist_path=plist_path)
        )
        return None


def find_fritz_api_key(fritz_info_path: pathlib.Path) -> Optional[str]:
    """Find API key from Fritz-Info.plist file.

    Args:
        fritz_info_path: Path to Fritz-Info.plist file.

    Returns: App API Key if it exists.
    """
    plist = read_plist(fritz_info_path) or {}
    if "apiKey" in plist:
        return plist["apiKey"]

    return None


def get_app_name(xcodeproj_path: pathlib.Path) -> Optional[str]:
    """Get app name from xcodeproj path.

    Args:
        xcodeproj_path: Path to xcodeproj folder.

    Returns: App Name if found.
    """
    cmd = (
        "xcodebuild  -project {xcodeproj} -showBuildSettings | grep PRODUCT_NAME | "
        'grep -Fv "FULL_PRODUCT_NAME"'
    )
    cmd = cmd.format(xcodeproj=str(xcodeproj_path))
    product_name_line = os.popen(cmd).read()

    if not product_name_line:
        return None

    return product_name_line.split("=")[1].strip()


def get_bundle_id(xcodeproj_path: pathlib.Path) -> Optional[str]:
    """Get bundle ID form xcodeproj.

    Args:
        xcodeproj_path: Path to xcodeproj folder.

    Returns: Bundle ID if found.
    """
    cmd = "xcodebuild -project {xcodeproj} -showBuildSettings | grep PRODUCT_BUNDLE_IDENTIFIER"
    cmd = cmd.format(xcodeproj=str(xcodeproj_path))
    product_bundle_line = os.popen(cmd).read()
    if not product_bundle_line:
        return None

    return product_bundle_line.split("=")[1].strip()


def create_fritz_info(path: pathlib.Path, app: fritz.App):
    """Create Fritz-Info.plist file at path.

    Args:
        path: Fritz-Info.plist path.
        app: App.
    """
    data = {
        "namespace": "production",
        "apiUrl": fritz.api_base.replace("client", "sdk"),
        "apiKey": app.api_key,
    }
    # TODO: Replace with non-deprecated method.
    # pylint: disable=deprecated-method
    plistlib.writePlist(data, str(path))


def add_path_to_xcodeproj(
    project: XcodeProject, relative_path: pathlib.Path
) -> XcodeProject:
    """Adds path to xcode project file.

    Args:
        project: Project to update.
        relative_path: Relative path from root of project.

    Returns: Updated Xcode project.
    """
    parent_name = relative_path.parent.name
    if not parent_name:
        project.add_file(str(relative_path))
        return project

    groups = project.get_groups_by_name(parent_name)
    if not groups:
        project.add_file(str(relative_path))
        return project

    project.add_file(str(relative_path), parent=groups[0])
    return project


def remove_file_in_xcodeproj(
    project: XcodeProject, filename: str
) -> XcodeProject:
    """Removes file from Xcode Project.

    Args:
        project: Project to modify.
        filename: Name of file to remove.

    Returns: Modified project.

    """
    files = project.get_files_by_name(filename)

    for file in files:
        # The remove_file_by_id function doesn't quite work, here are some
        # modifications to successfully remove the file.
        project.remove_file_by_id(file.get_id())

        for group in project.objects.get_objects_in_section(u"PBXGroup"):
            if file.get_id() in group.children:
                group.remove_child(file)

        for build_file in project.objects.get_objects_in_section(
            u"PBXBuildFile"
        ):
            if build_file.fileRef == file.get_id():
                del project.objects[build_file.get_id()]

        for build_file in project.objects.get_objects_in_section(
            "PBXFileReference"
        ):
            if build_file.get_id() == file.get_id():
                del project.objects[build_file.get_id()]

    return project



"""
cli.app.code_snippet

Base class for CodeSnippets.
"""

from fritz import errors


class CodeSnippet(object):
    """Defines a block of code to update. """

    def __init__(self, app_config, app):
        self.app_config = app_config
        self.app = app

    def __str__(self):
        return self.__class__.__name__

    @property
    def is_added(self):
        """Returns True if code snippet has been added."""
        raise NotImplementedError

    @property
    def is_updated(self):
        """Returns True if added code snippet is up to date."""
        # override if can be updated.
        return True

    def add(self):
        """Add code snippet."""
        raise NotImplementedError

    def update(self):
        """Update existing code snippet."""


class SnippetExecutionError(errors.FritzError):
    """Error executing snippet."""



"""
fritz.api_resources.model
~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""

import fritz
from fritz.api_resources import fritz_object


class Model(fritz_object.FritzObject):
    """Fritz Model object."""

    OBJECT_NAME = "Model"

    def summary(self):
        """Print summary of object."""
        details = [
            ("Created At", self.created_date),
            ("Description", self.description),
            ("Model Format", self.model_format),
            ("Project ID", self.project_id),
            ("Active Version", self.active_version),
            ("Is Global", self.is_global),
        ]
        return self._summary(self.name, details)

    @property
    def framework(self):
        """Gets associated framework for model.

        Returns: frameworks.ModelFramework
        """
        if self.model_format == "coreml":
            return fritz.frameworks.CORE_ML
        if self.model_format == "tflite":
            return fritz.frameworks.TENSORFLOW_LITE
        if self.model_format == "tfmobile":
            return fritz.frameworks.TENSORFLOW_MOBILE
        if self.model_format == "keras":
            return fritz.frameworks.KERAS

        return None

    @classmethod
    def get(cls, api_key=None, model_id=None):
        """Get model.

        Args:
            model_id: ID of model to query.
            api_key: Optional API Key to use.

        Returns: Model
        """
        if not model_id:
            raise fritz.errors.ArgumentRequiredError("model_id")

        url = "/model/" + model_id
        response = cls.client_get(url, api_key=api_key)
        return fritz.utils.convert_to_fritz_object(response)

    @classmethod
    def list(cls, api_key=None, project_id=None, all_projects=False):
        """Get model, restricting results to specified arguments.

        Args:
            api_key: Optional API Key to use.
            project_id: Project ID
            all_projects: If True and no project_id specified, returns models
                from all projects. Default False.

        Returns: List[Model]
        """
        url = "/model"
        if not project_id and not all_projects:
            project_id = fritz.project_id

        params = {"project_uid": project_id}
        response = cls.client_get(url, api_key=api_key, params=params)

        return [
            fritz.utils.convert_to_fritz_object(model)
            for model in response["models"]
        ]

    def get_version(self, version_number=None):
        """Get specific version matching version number.

        Args:
            version_number (int): Version to fetch. If not provided uses
                active_version.

        returns: Optional[fritz.ModelVersion]
        """
        version_number = version_number or self.active_version
        versions = [
            version
            for version in fritz.ModelVersion.list(self.id)
            if version.version_number == version_number
        ]

        if versions:
            return versions[0]

        return None

    def download(self, api_key=None, version_number=None, output_dir=None):
        """Download model file of either active version or specified version.

        Args:
            api_key (str): Optional API Key to use.
            version_number (int): Optional version number. If not specified
                downloads active version.
            output_dir (str): Output directory.

        Returns: Tuple of (ModelVersion, path) of downloaded file.
        """

        version_number = version_number or self.active_version

        versions = fritz.ModelVersion.list(self.id, api_key=api_key)
        version = next(
            (
                version
                for version in versions
                if version.version_number == version_number
            ),
            None,
        )
        if not version:
            return (None, None)

        path = version.download(download_dir=output_dir)
        return (version, path)

    def update(self, api_key=None, **updates):
        """Update model with specified updates.

        Args:
            api_key: Optional API Key to use.
            kwargs: Dictionary of updates.

        Returns: New model with updates.
        """
        url = "/model/" + self.id
        result = self.client_put(url, params=updates)
        return fritz.utils.convert_to_fritz_object(result)



"""
cli.app.android.android_config

Android Configuration.
"""

from typing import Optional
import pathlib
from fritz import errors
from cli.app.android import android_helpers


def _find_android_manifest(root: pathlib.Path) -> pathlib.Path:
    paths = root.rglob("**/src/main/AndroidManifest.xml")
    return next(paths, None)


def _find_app_gradle_path(root: pathlib.Path) -> pathlib.Path:
    paths = root.rglob("build.gradle")
    for path in paths:
        if "com.android.application" in path.read_text():
            return path

    return None


def _find_project_gradle_path(root: pathlib.Path) -> Optional[pathlib.Path]:
    gradle_path = root / "build.gradle"
    if not gradle_path.exists():
        return None
    return gradle_path


class AndroidConfig(object):
    """Configuration of Android app project."""

    def __init__(
        self,
        root: pathlib.Path,
        android_manifest_path=None,
        app_gradle_path=None,
        project_gradle_path=None,
    ):
        self.root = root
        self._android_manifest_path = android_manifest_path
        self._app_gradle_path = app_gradle_path
        self._project_gradle_path = project_gradle_path

    @property
    def platform(self):
        """Platform of app."""
        return "android"

    @property
    def app_name(self) -> Optional[str]:
        """App Name. """
        return android_helpers.find_app_name(
            self.root, self.android_manifest_path
        )

    @property
    def package_name(self) -> Optional[str]:
        """Returns APK ID for project."""
        app_id = android_helpers.find_application_id(self.app_gradle_path)
        if app_id:
            return app_id
        return android_helpers.find_package_name(self.android_manifest_path)

    @property
    def android_manifest_path(self) -> Optional[pathlib.Path]:
        """Path to Android manifest."""
        return self._android_manifest_path

    @property
    def app_gradle_path(self) -> Optional[pathlib.Path]:
        """Path to app build.gradle path."""
        return self._app_gradle_path

    @property
    def project_gradle_path(self) -> Optional[pathlib.Path]:
        """Path to project build.gradle path."""
        return self._project_gradle_path

    @property
    def main_activity_path(self) -> Optional[pathlib.Path]:
        """Path to main activity."""
        main_activity_name = android_helpers.find_main_activity(
            self.android_manifest_path
        )
        paths = self.root.rglob("**/" + main_activity_name + ".java")
        return next(paths, None)

    @property
    def fritz_api_key(self) -> Optional[pathlib.Path]:
        """Fritz API Key found in app. """
        return android_helpers.find_fritz_api_key(self.app_gradle_path)

    @classmethod
    def load(cls, root: pathlib.Path):
        """Search root path and initialize config with found files.

        Args:
            root: Root path of Android project.

        Returns: AndroidConfig.
        """
        config = cls(
            root,
            android_manifest_path=_find_android_manifest(root),
            app_gradle_path=_find_app_gradle_path(root),
            project_gradle_path=_find_project_gradle_path(root),
        )

        if not config.android_manifest_path:
            raise errors.MissingProjectFileError("AndroidManifest.xml")

        if not config.app_gradle_path:
            raise errors.MissingProjectFileError("app/build.gradle")

        if not config.project_gradle_path:
            raise errors.MissingProjectFileError("Project build.gradle")

        return config

    def summary(self):
        """Summary of Android config."""
        return """
App Name:            {app_name}
Package Name:        {package_name}
Android Manifest:    {manifest}
App Gradle Path:     {app_gradle}
Project Gradle Path: {project_gradle}
        """.format(
            app_name=self.app_name,
            package_name=self.package_name,
            manifest=self.android_manifest_path,
            app_gradle=self.app_gradle_path,
            project_gradle=self.project_gradle_path,
        )



"""
This module contains the code for setting up an Android app.
"""
import fritz
from cli.app.ios.ios_config import IOSConfig
from cli.app.ios import ios_snippets


def code_snippets(app_config: IOSConfig, app: fritz.App):
    """Load all code snippets.

    Args:
        app_config: iOS config
        app: App to connect Android project to.

    Returns: List of code snippets.
    """
    snippets = [
        ios_snippets.CreateFritzInfo,
        ios_snippets.FritzInfoXcodeProject,
        ios_snippets.FritzInfoAPIKeyUpdater,
        ios_snippets.AddFritzImport,
        ios_snippets.AddFritzConfigure,
        ios_snippets.AddPodfile,
        ios_snippets.AddFritzToPodfile,
    ]
    return [snippet(app_config, app) for snippet in snippets]



"""
fritz.api_resources
~~~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""

from fritz.api_resources.model import Model
from fritz.api_resources.model_version import ModelVersion
from fritz.api_resources.model_snapshot import ModelSnapshot
from fritz.api_resources.model_grade_report import ModelGradeReport
from fritz.api_resources.project import Project
from fritz.api_resources.app import App



"""
fritz.api_resources.model_snapshot
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
import logging

import fritz
from fritz.api_resources.fritz_object import FritzObject

# pylint: disable=invalid-name
_logger = logging.getLogger(__name__)


class ModelSnapshot(FritzObject):
    """A collection of models built from the same training model. """

    OBJECT_NAME = "ModelSnapshot"

    # pylint: disable=duplicate-code
    @classmethod
    def create(
        cls,
        api_key=None,
        project_id=None,
        keras_model=None,
        converters=None,
        model_ids=None,
        filename=None,
        set_active=None,
        metadata=None,
    ):
        """Create new ModelSnapshot from a Keras model.

        Calling create will run all provided converters on `keras_model`
        and upload them to the API. All models will be bundled into
        the same ModelSnapshot

        Args:
            api_key (Optional[str]): API Key. Optional if `fritz.configure`
                was called.
            project_id (Optional[str]): Project ID. Optional if
                `fritz.configure` was called with project_id.
            keras_model (keras.models.Model): Keras Model.
            converters (Dict[frameworks.ModelFramework,(keras.model.Model) ->
                Any]): Dictionary mapping model framework to conversion
                function.
            model_ids (Dict[`frameworks.ModelFramework`,str]): Dictionary
                mapping model framework to model ids.  If model_id not set
                for a given platform, a new model will be created.
            filename (str): Name of Keras model output filename.
            set_active (bool): If True, model will be set as the active version
               in Fritz. If it is True, any devices runninng this model will
               download the latest version.
            metadata (dict): Dictionary of metadata.

        Returns: Tuple[fritz.ModelSnapshot, List[fritz.ModelVersion],
            List[fritz.Model]]
        """
        keras_file = fritz.frameworks.KerasFile(keras_model)
        model_ids = model_ids or {}
        versions = []
        models = []

        keras_model_id = model_ids.get(fritz.frameworks.KERAS)
        version, snapshot, model = fritz.ModelVersion.create(
            api_key=api_key,
            project_id=project_id,
            filename=filename,
            data=keras_file.to_bytes(),
            model_id=keras_model_id,
            set_active=set_active,
            metadata=metadata,
        )
        versions.append(version)
        models.append(model)

        for framework, converter in converters.items():
            _logger.info("Converting %s model", framework.name)

            converted_model = converter(keras_model)
            framework_file = framework.to_file(converted_model)

            version, snapshot, _ = fritz.ModelVersion.create(
                api_key=api_key,
                project_id=project_id,
                model_id=model_ids.get(framework),
                snapshot_id=snapshot.id,
                filename=framework.as_framework_filename(filename),
                data=framework_file.to_bytes(),
                set_active=set_active,
                metadata=metadata,
            )
            versions.append(version)

        return snapshot, versions, models



"""
fritz.train
~~~~~~~~~~~

:copyright: © 2018 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""

from fritz.train.model_checkpoint import FritzSnapshotCallback

assert FritzSnapshotCallback



"""
fritz.utils
~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
import fritz


def _get_object_classes():
    """Get object classes by object name."""
    return {
        fritz.ModelVersion.OBJECT_NAME: fritz.ModelVersion,
        fritz.ModelSnapshot.OBJECT_NAME: fritz.ModelSnapshot,
        fritz.Model.OBJECT_NAME: fritz.Model,
        fritz.App.OBJECT_NAME: fritz.App,
        fritz.Project.OBJECT_NAME: fritz.Project,
        fritz.ModelGradeReport.OBJECT_NAME: fritz.ModelGradeReport,
    }


def convert_to_fritz_object(response_data):
    """Convert response data to a corresponding FritzObject

    Args:
        response_data (dict): Response data from request.

    Returns:
        FritzObject: converted dict into a fritz object.
    """
    if not isinstance(response_data, dict):
        return response_data

    for key in response_data:
        response_data[key] = convert_to_fritz_object(response_data[key])

    object_classes = _get_object_classes()
    if "object_name" in response_data:
        return object_classes[response_data["object_name"]](**response_data)

    return response_data



"""
cli.app.ios.ios_snippets

Code Snippets for setting up Fritz with a Swift iOS app.
"""
# pylint: disable=abstract-method

import os
import pathlib

from pbxproj import XcodeProject

from cli.app import code_snippet
from cli.app.ios import ios_helpers
from cli.app import file_manager


class FritzInfoSnippet(code_snippet.CodeSnippet):
    """Base Snippet for Fritz-Info.plist files."""

    def __init__(self, config, app):
        super().__init__(config, app)
        self.path = config.fritz_info_path or config.default_fritz_info_path


class AppDelegateSnippet(code_snippet.CodeSnippet):
    """Base Snippet for AppDelegate.swift files."""

    def __init__(self, config, app):
        super().__init__(config, app)
        self.path = config.app_delegate_path


class CreateFritzInfo(FritzInfoSnippet):
    """Handles create and renaming Fritz-Info.plist file.

    This does not add to the Xcode project, just creates the file.
    """

    @property
    def is_added(self):
        found = next(self.app_config.root.rglob("Fritz-Info.plist"), None)
        return found is not None

    @property
    def is_updated(self):
        if not self.is_added:
            return True

        # If the path specified does not match the one in the project,
        # this will return false.
        return self.path.exists()

    def add(self):
        ios_helpers.create_fritz_info(self.path, self.app)

    def update(self):
        existing = next(self.app_config.root.rglob("Fritz-Info.plist"))
        existing.rename(self.path)


class FritzInfoAPIKeyUpdater(FritzInfoSnippet):
    """Handles updating the API Key if it does not match."""

    @property
    def is_added(self):
        # This will never add an api key... only will update existing ones.
        return True

    @property
    def is_updated(self):
        if not self.is_added:
            return True

        existing_path = next(
            self.app_config.root.rglob("Fritz-Info.plist"), None
        )
        if not existing_path:
            return True

        return self.app.api_key == ios_helpers.find_fritz_api_key(
            existing_path
        )

    def add(self):
        pass

    def update(self):
        existing_path = next(self.app_config.root.rglob("Fritz-Info.plist"))
        ios_helpers.create_fritz_info(existing_path, self.app)


class FritzInfoXcodeProject(FritzInfoSnippet):
    """Add Fritz Info file to Xcode Project."""

    @property
    def project(self):
        """Xcode Project."""
        project_file = "{xcode_proj_file}/project.pbxproj".format(
            xcode_proj_file=str(self.app_config.xcodeproj_path)
        )
        return XcodeProject.load(project_file)

    @property
    def is_added(self):
        return len(self.project.get_files_by_name("Fritz-Info.plist")) > 0

    @property
    def is_updated(self):
        if not self.is_added:
            return True

        relative_path = self.path.relative_to(self.app_config.root)
        # If there are files matching the path, then file exists in the correct
        # place so no need to move.
        return len(self.project.get_files_by_path(str(relative_path))) > 0

    def add(self):
        relative_path = self.path.relative_to(self.app_config.root)
        project = ios_helpers.add_path_to_xcodeproj(
            self.project, relative_path
        )
        project.save()

    def update(self):
        relative_path = self.path.relative_to(self.app_config.root)
        project = ios_helpers.remove_file_in_xcodeproj(
            self.project, relative_path.name
        )
        project = ios_helpers.add_path_to_xcodeproj(project, relative_path)
        project.save()


class AddFritzImport(AppDelegateSnippet):
    """Add Fritz import in AppDelegate."""

    @property
    def is_added(self):
        return "import Fritz" in self.path.read_text()

    def add(self):
        text = file_manager.insert_before(
            self.path, "@UIApplicationMain", "import Fritz\n"
        )
        self.path.write_text(text)


class AddFritzConfigure(AppDelegateSnippet):
    """Add Fritz configure call in app-delegate."""

    @property
    def is_added(self):
        return "FritzCore.configure" in self.path.read_text()

    def add(self):
        hint = "didFinishLaunchingWithOptions"
        text = file_manager.insert_in_function_after_hint(
            self.path, hint, "FritzCore.configure()"
        )
        self.path.write_text(text)


class PodfileSnippet(code_snippet.CodeSnippet):
    """Base Snippet for Podfile."""

    def __init__(self, config, app):
        super().__init__(config, app)
        self.path = config.podfile_path or config.default_podfile_path


class AddPodfile(PodfileSnippet):
    """Creates Podfile."""

    @property
    def is_added(self):
        return self.path.exists()

    def add(self):
        os.popen("pod init " + str(self.app_config.xcodeproj_path)).read()
        pod_path = pathlib.Path.cwd() / "Podfile"
        pod_path.rename(self.path)
        self.app_config.podfile_path = self.path


class AddFritzToPodfile(PodfileSnippet):
    """Adds Fritz dependency to Podfile."""

    @property
    def is_added(self):
        if not self.path.exists():
            return False

        return "Fritz" in self.path.read_text()

    def add(self):
        # Line that appears in default podfile.
        fritz_install_line = "  pod 'Fritz'\n"

        text = self.path.read_text()
        target = self.app_config.app_name
        default_target_line = "# Pods for {target}".format(target=target)
        if default_target_line in text:
            text = file_manager.insert_after(
                self.path, default_target_line, fritz_install_line
            )
            self.path.write_text(text)
            return

        pod_target_line = "target '{target}'".format(
            target=self.app_config.app_name
        )
        if pod_target_line in text:
            text = file_manager.insert_after(
                self.path, pod_target_line, fritz_install_line
            )
            self.path.write_text(text)
            return

        raise code_snippet.SnippetExecutionError(
            "Could not determine where to insert Fritz pod."
        )



"""
cli.app.android.android_helpers

Helper methods for parsing various values from Android project.
"""

from typing import Optional
import re
from xml.etree import ElementTree
import pathlib


def _get_app_label(android_manifest_path: pathlib.Path) -> str:
    tree = ElementTree.parse(str(android_manifest_path))
    root = tree.getroot()
    application_node = root.find("application")
    app_label = application_node.get(
        "{http://schemas.android.com/apk/res/android}label"
    )
    return app_label


def _get_app_name_from_strings(
    root: pathlib.Path, app_label: str
) -> Optional[str]:

    label_key = app_label.split("@string/")[1]
    paths = root.rglob("**/src/main/res/values/strings.xml")

    for string_path in paths:
        tree_string = ElementTree.parse(str(string_path))
        root_string = tree_string.getroot()
        string_nodes = root_string.findall("string")

        for string_node in string_nodes:
            if string_node.get("name") == label_key:
                return string_node.text

    return None


def find_app_name(
    root: pathlib.Path, android_manifest_path: pathlib.Path
) -> Optional[str]:
    """Parses app name from a given android manifest file.

    Args:
        root: Root path of project.
        android_manifest_path: Path to AndroidManifest.xml file.

    Returns: App Name if found, None otherwise.
    """
    app_label = _get_app_label(android_manifest_path)
    if "@string/" not in app_label:
        return app_label

    return _get_app_name_from_strings(root, app_label)


def find_package_name(android_manifest_path: pathlib.Path) -> str:
    """Find the name of the package.

    Returns:
        package name
    """
    tree = ElementTree.parse(str(android_manifest_path))
    root = tree.getroot()
    package_name = root.get("package")
    return package_name


def find_application_id(build_path: pathlib.Path) -> str:
    """Find the application id in app/build.gradle.

    Args:
        build_path: the path to build.gradle

    Returns:
        the applicationId
    """
    for line in build_path.open():
        if "applicationId" not in line:
            continue
        matched_items = re.search(r'applicationId "(.+)"', line)
        if matched_items:
            return matched_items.groups(0)[0]

    return None


def find_fritz_api_key(main_activity_path: pathlib.Path) -> str:
    """Find the apk id.

    Returns:
        the apk id.
    """
    for line in main_activity_path.open():
        if "Fritz" not in line:
            continue
        matched_items = re.search(r'Fritz.configure\(this, "(.+)"\);', line)
        if matched_items:
            return matched_items.groups(0)[0]

    return None


def find_main_activity(android_manifest_path: pathlib.Path):
    """Find the 'MainActivity'

    Returns:
        the name of the class
    """
    tree = ElementTree.parse(str(android_manifest_path))
    root = tree.getroot()
    application_node = root.find("application")
    for activity_node in application_node.findall("activity"):
        intent_filter = activity_node.find("intent-filter")
        if intent_filter is None:
            continue
        action_node = intent_filter.find("action")
        if action_node is None:
            continue
        action_name = action_node.get(
            "{http://schemas.android.com/apk/res/android}name"
        )
        if action_name == "android.intent.action.MAIN":
            class_name = activity_node.get(
                "{http://schemas.android.com/apk/res/android}name"
            )
            return class_name.split(".")[-1]

    return None



"""Helpers for printing in the cli."""
import click


CODE_COLOR = "yellow"
IMPORTANT_COLOR = "green"
ACK_COLOR = "yellow"
ERROR_COLOR = "red"


def code(code):
    """Show a code example."""
    click.echo(click.style(code, fg=CODE_COLOR))


def ack(content):
    """Create acknowledged text after a user action."""
    click.echo(click.style(content, fg=ACK_COLOR))


def message(content):
    """Create acknowledged text after a user action."""
    click.echo(content)


def notice(content):
    """Print notice describing summary of section. """
    click.echo(click.style(content, fg=IMPORTANT_COLOR))


def alert(content):
    """Print alert. """
    click.echo(click.style(content, fg=ERROR_COLOR))


def error(content, fix_content=None):
    """Create an error message with optional content to fix the issue."""
    click.echo(click.style("[ERROR]: " + content, fg=ERROR_COLOR))
    if fix_content:
        click.echo(click.style(fix_content, fg=ERROR_COLOR))


def warn(content, fix_content=None):
    """Create a warning message with optional content to fix the issue."""
    click.echo(click.style("[WARNING]: " + content, fg="yellow"))
    if fix_content:
        click.echo(click.style(fix_content, fg="yellow"))


def title(content):
    """Create a title in the terminal."""
    click.echo(click.style("\n--------------------------", fg=IMPORTANT_COLOR))
    click.echo(click.style(content, fg=IMPORTANT_COLOR))
    click.echo(click.style("--------------------------", fg=IMPORTANT_COLOR))


def field(name, value):
    """Print a field.

    Args:
        name (str): Name of field.
        value (str): Value.
    """
    line = "  {name:20}: {value}".format(name=name, value=value)
    click.echo(click.style(line, fg=IMPORTANT_COLOR))


def header(content):
    """Print header with newline before and dashes after.

    Args:
        content (str): Content of header.
    """
    click.echo("\n" + content)
    click.echo("-" * len(content))


def formatted_table(formats, rows):
    """Print formatted table.

    Args:
        formats (list[tuple]): Tuple of (key, format, header_name)
        rows (List[Dict]): Data rows.
    """
    if not rows:
        return

    format_str = " | ".join(
        column_fmt % max(*[len(str(row[key])) for row in rows], len(header))
        for key, column_fmt, header in formats
    )
    header_values = {key: value for key, _, value in formats}
    header = format_str.format(**header_values)

    click.echo(header)
    click.echo("-" * len(header))

    for row in rows:
        click.echo(format_str.format(**row))

    click.echo("")


def make_plural(count, word):
    """Makes world plural if more than one item.

    Args:
        count (int): count of items.
        word (str): Word.

    Returns: Modified word.
    """
    if count > 1:
        return word + "s"

    return word



"""
Fritz is a library to simplify your mobile machine learning workflow.

usage:
   >>> import fritz
   >>> fritz.configure(api_key="<api_key>", project_id="<project_id>")
   >>> version = fritz.ModelVersion.get("<model_version_id>")
   >>> benchmark = version.benchmark()
   >>> benchmark.summary()
       ------------------------
       Fritz Model Grade Report
       ------------------------
       Core ML Compatible:              True
       Predicted Runtime (iPhone X):    31.4 ms (31.9 fps)
       Total MFLOPS:                    686.90
       Total Parameters:                1,258,580


:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
# pylint: disable=invalid-name
from fritz.api_resources import *  # noqa

from fritz import frameworks  # noqa

# Configuration Variables
api_key = None
project_id = None
api_base = "https://api.fritz.ai/client/v1"


def configure(**kwargs):
    """Sets Fritz configuration variables.

    If no variables are passed, the function will attempt to load configuration
    variables from a ~/.fritz file.

    To configure this file, run `fritz config` from the command line.

    Args:
        api_key (str): Client API Key used to authenticate requests to Fritz.
        project_id (str): Project ID to store models in.
        api_base (str): Base URL of Fritz API
    """
    # pylint: disable=global-statement
    global api_key, project_id, api_base
    if not kwargs:
        # If no kwargs are passed, attempt to initialize from ~/.fritz file.
        from fritz import configuration

        configuration.init_config()
        return

    api_key = kwargs.get("api_key") or api_key
    project_id = kwargs.get("project_id") or project_id
    api_base = kwargs.get("api_base") or api_base


def reset_configuration():
    """Resets Fritz configuration variables."""
    # pylint: disable=global-statement
    global api_key, project_id, api_base
    api_key = None
    project_id = None
    api_base = None



"""
cli.app.app_setup
~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
from typing import Optional

import pathlib
import click

import fritz
from cli.app import app_files
from cli.app import android
from cli.app import ios
from cli import fritz_object_selector
from cli import print_utils


@click.command()
@click.argument(
    "root", required=False, type=pathlib.Path, default=str(pathlib.Path.cwd())
)
def setup(root):
    """Connecting an existing mobile app to Fritz.

    Args:
        root: root of app directory.
    """
    android_files = list(root.rglob("AndroidManifest.xml"))
    ios_files = list(root.rglob("Info.plist"))
    if android_files and ios_files:
        raise fritz.errors.FritzError(
            "Found iOS and Android projects. Please specify a directory"
            " with only an iOS or Android project"
        )
    if android_files:
        app_config = android.AndroidConfig.load(root)
        platform = "Android"
        build_snippets = android.code_snippets
    elif ios_files:
        platform = "iOS"
        app_config = ios.IOSConfig.load(root)
        build_snippets = ios.code_snippets

    print_utils.ack(
        "Connecting an {platform} app to Fritz.\n".format(platform=platform)
    )
    print_utils.message(
        "Loading {platform} configuration from {root}\n".format(
            platform=platform, root=root
        )
    )
    print_utils.notice("Successfully loaded app configuration")
    print_utils.notice(app_config.summary())

    app = get_fritz_app(app_config)
    code_snippets = build_snippets(app_config, app)

    app_files.update(code_snippets)


def _get_app_from_config(app_config) -> Optional[fritz.App]:
    """Load app from app_config.

    Args:
        app_config: App configuration.

    Returns: App matching current config or None
    """
    apps = fritz.App.list(
        apk_id=app_config.package_name,
        platform=app_config.platform,
        app_api_key=app_config.fritz_api_key,
        all_projects=True,
    )
    if not apps:
        return None

    if len(apps) > 1:
        print_utils.warn("Found more than one app, taking first")

    return apps[0]


def _choose_app_from_matching_apk_id(app_config) -> Optional[fritz.App]:
    """Choose an app that matches the APK ID from any project.

    When creating an app, the apk may exist in any of your projects. If it does
    this gives you the opportunity to an existing app, rather than creating a
    new one.

    Args:
        app_config: App configuration.
    """
    apps = fritz.App.list(
        apk_id=app_config.package_name,
        platform=app_config.platform,
        all_projects=True,
    )

    if not apps:
        return None

    app = fritz_object_selector.choose_app(apps, include_none=True)
    return app


def _configure_app(details: dict) -> dict:
    """Configure values from data needed to create an app.

    Args:
        details: Dict of values to pass to app creation.

    Returns: Updated details.
    """
    project = fritz.Project.get(project_id=details["project_id"])
    summary = {
        "App Name": details["app_name"],
        "APK ID": details["apk_id"],
        "Project Name": project.name,
    }

    lines = []
    max_name = max(len(name) for name, _ in summary.items())
    for name, value in summary.items():
        row_fmt = "  {name:%s}: {value}" % (max_name + 2)
        lines.append(row_fmt.format(name=name, value=value))

    summary = "\n".join(lines)
    print_utils.ack("Your app will be created with the following details:\n")
    print_utils.notice("App Summary")
    print_utils.notice(summary)

    if click.confirm("\nModify App Name or Project?", default=False):
        details["app_name"] = click.prompt(
            "App Name", default=details["app_name"], show_default=True
        )
        projects = fritz.Project.list()

        if len(projects) > 1 and click.confirm(
            "Choose a different project?", default=False
        ):
            project = fritz_object_selector.choose_project(projects)
            details["project_id"] = project.id

        return _configure_app(details)

    return details


def _create_app(app_config) -> fritz.App:
    """Create fritz app from app config.

    Args:
        app_config: App configuration.

    Returns: Created Fritz app.
    """
    project = fritz.Project.get()
    details = {
        "app_name": app_config.app_name,
        "apk_id": app_config.package_name,
        "platform": app_config.platform,
        "project_id": project.id,
    }

    _configure_app(details)

    return fritz.App.create(
        app_name=details["app_name"],
        apk_id=details["apk_id"],
        platform=details["platform"],
        project_id=details["project_id"],
    )


def get_fritz_app(app_config) -> fritz.App:
    """Steps needed to choose the fritz app to connect to project.

    Args:
        app_config: App configuration

    Returns: App that will be used for setup.
    """
    app = None
    if app_config.fritz_api_key:
        app = _get_app_from_config(app_config)

    if not app and app_config.fritz_api_key:
        print_utils.warn(
            "Fritz API Key found, but no app with matching "
            "API Key found connected to your account."
        )

    if not app:
        print_utils.message("Checking for other apps with matching apk id.\n")
        app = _choose_app_from_matching_apk_id(app_config)
        if app:
            return app

    if not app:
        print_utils.ack("\nNo matching apps found.\n")
        app = _create_app(app_config)

    return app



"""
This module contains the code for setting up an Android app.
"""
import fritz
from cli.app.android.android_config import AndroidConfig
from cli.app.android import android_snippets


def code_snippets(app_config: AndroidConfig, app: fritz.App):
    """Load all code snippets.

    Args:
        app_config: Android config
        app: App to connect Android project to.

    Returns: List of code snippets.
    """
    snippets = [
        android_snippets.FritzCustomModelServiceSnippet,
        android_snippets.InternetPermissionsSnippet,
        android_snippets.ImportFritzSnippet,
        android_snippets.ConfigureFritzSnippet,
        android_snippets.AddBuildDependencies,
        android_snippets.AddProjectBuildDependencies,
        android_snippets.AddRenderscript,
    ]
    return [snippet(app_config, app) for snippet in snippets]



"""Manages app project files that should be edited."""
import pathlib


def insert_before(path, value_to_find, code_to_insert) -> str:
    """Insert code before any occurrence of value.

    Args:
        path: Path to file
        value_to_find: search term.
        code_to_insert: Text to insert before search term.

    Returns: Text of modified file.
    """
    contents = []
    for line in path.open(mode="r"):
        if value_to_find not in line:
            contents.append(line)
            continue

        contents.append(code_to_insert)
        contents.append(line)

    return "".join(contents)


def insert_after(path, value_to_find, code_to_insert):
    """Insert code after any first of value.

    Args:
        path: Path to file
        value_to_find: search term.
        code_to_insert: Text to insert before search term.

    Returns: Text of modified file.
    """
    content = []
    added = False
    for line in path.open():
        content.append(line)
        if value_to_find in line and not added:
            content.append(code_to_insert)
            added = True

    return "".join(content)


def insert_line_at_block_end(path, block_start, line_to_insert):
    """Insert code at the end of a text block.

    Args:
        path: Path to file
        block_start: Search term that triggers the beginning of block.
        line_to_insert: Text to insert at end of block.

    Returns: Text of modified file.
    """
    contents = path.open().readlines()
    start, end = find_block_indexes(contents, block_start)

    if start and end and start != end:
        contents.insert(end, line_to_insert + "\n")

    return "".join(contents)


def add_indentation(num_indents, num_spaces=4):
    """Add indentation

    Args:
        num_indents: number of indentations
        num_spaces: Num spaces per indent. Defaults to 4.

    Returns:
        spaces
    """
    return " " * num_indents * num_spaces


def find_block_indexes(lines, key):
    """Find indices of opening and closing block.

    Note, this does not yet handle recursive blocks.

    For example:
    1  if statement1 {
    2    if statement2 {
    3    }
    4  }

    will return 1, 3.

    Args:
        lines: List of text lines.
        key: Search term which starts block.

    Returns: tuple of start and end of block.
    """
    total_lines = len(lines)
    start = None
    end = None
    for line_num in range(total_lines):
        line = lines[line_num]
        if key in line:
            for num in range(line_num, total_lines):
                line = lines[num]
                if "{" in line:
                    start = num
                if "}" in line:
                    end = num
                    break
            break
    return start, end


def insert_in_function_after_hint(
    path: pathlib.Path,
    function_signature_hint: str,
    code: str,
    indentation: int = 4,
):
    """Insert code in function with signature hint text in definition.

    Args:
        path: Path to code.
        function_signature_hint: text to look for in function signature.
            Should be unique.
        code: Code to insert.
        indentation: Width of indentation.

    Returns: Modified file.
    """
    found_function = False
    in_function = False
    output_lines = []
    configure_added = False
    for line in path.open():
        output_lines.append(line)

        if function_signature_hint in line:
            found_function = True

        # Searching for start of function
        if found_function and not in_function:
            if "{" in line:
                in_function = True
            continue

        if in_function and "}" in line:
            in_function = False
            continue

        if not in_function:
            continue

        if in_function and not configure_added:
            # Adding FritCore.configure
            message = "// Automatically added by fritz app setup."
            output_lines.append(" " * indentation + message + "\n")
            output_lines.append(" " * indentation + code + "\n")
            configure_added = True

    return "".join(output_lines)



"""
cli.app.app_files

Methods to help setup Android project with Fritz App.
"""
from typing import List
import click
from cli import print_utils
from cli.app import code_snippet


def update(snippets: List[code_snippet.CodeSnippet]):
    """Update all code snippets to connect to Fritz.

    Args:
        snippets: List of CodeSnippets to update.
    """
    up_to_date = [
        snippet
        for snippet in snippets
        if snippet.is_added and snippet.is_updated
    ]
    missing = [snippet for snippet in snippets if not snippet.is_added]
    needs_update = [snippet for snippet in snippets if not snippet.is_updated]

    print_utils.message("Updating code snippets with latest configuration.")

    if up_to_date:
        print_utils.notice("\nThe following snippets are up to date:")
        for snippet in up_to_date:
            print_utils.notice("  - " + str(snippet))

    if needs_update:
        print_utils.notice("\nThe following snippets will be updated:")
        for snippet in needs_update:
            print_utils.ack("  - " + str(snippet))

    if missing:
        print_utils.ack("\nThe following snippets will be added:")
        for snippet in missing:
            print_utils.ack("  - " + str(snippet))
        print_utils.ack("")

    if not click.confirm("Continue updating app files?", default=True):
        return

    errors = []

    for snippet in missing:
        try:
            snippet.add()
            print_utils.notice("Successfully added code in " + str(snippet))
        except code_snippet.SnippetExecutionError as err:
            print_utils.error(str(err))
            errors.append(err)

    for snippet in needs_update:
        try:
            snippet.update()
            print_utils.notice("Successfully added code in " + str(snippet))
        except code_snippet.SnippetExecutionError as err:
            print_utils.error(str(err))
            errors.append(err)

    if errors:
        raise code_snippet.SnippetExecutionError(
            "Some code changes failed to apply properly"
        )



"""
cli.app.android.android_config

Android Configuration.
"""

from typing import Optional
import pathlib
from fritz import errors

from cli.app.ios import ios_helpers


def _find_xcodeproj_path(root: pathlib.Path) -> pathlib.Path:
    paths = root.rglob("**/*.xcodeproj")
    return next(paths, None)


def _find_app_delegate_path(root: pathlib.Path) -> pathlib.Path:
    paths = root.rglob("AppDelegate.swift")
    return next(paths, None)


def _find_fritz_info_path(root: pathlib.Path) -> Optional[pathlib.Path]:
    paths = root.rglob("Fritz-Info.plist")
    return next(paths, None)


def _find_podfile_path(root: pathlib.Path) -> Optional[pathlib.Path]:
    paths = root.rglob("Podfile")
    return next(paths, None)


class IOSConfig(object):
    """Configuration of IOS app project."""

    def __init__(
        self,
        root: pathlib.Path,
        app_delegate_path=None,
        xcodeproj_path=None,
        fritz_info_path=None,
        podfile_path=None,
    ):
        self.root = root
        self._xcodeproj_path = xcodeproj_path
        self._app_delegate_path = app_delegate_path
        self._fritz_info_path = fritz_info_path
        self._podfile_path = podfile_path

        # Loading the app name and bundle name is expensive.
        # allow this to be loaded once on init.
        self._app_name = ios_helpers.get_app_name(self.xcodeproj_path)
        self._package_name = ios_helpers.get_bundle_id(self.xcodeproj_path)

    @property
    def platform(self):
        """Platform of app."""
        return "ios"

    @property
    def app_name(self) -> Optional[str]:
        """App Name. """
        return self._app_name

    @property
    def package_name(self) -> Optional[str]:
        """Returns APK ID for project."""
        return self._package_name

    @property
    def app_delegate_path(self) -> Optional[pathlib.Path]:
        """Path to AppDelegate."""
        return self._app_delegate_path

    @property
    def default_podfile_path(self) -> pathlib.Path:
        """Default path to Podfile."""
        return self.root / "Podfile"

    @property
    def default_fritz_info_path(self) -> pathlib.Path:
        """Gets path to info file by searching common directory structures."""
        supporting_files_folder = self.root.rglob("Supporting Files/")
        for folder in supporting_files_folder:
            # only want to look at directories.
            if not folder.is_dir():
                continue

            # Do not want to use the supporting files folder from a pod.
            if "Pods" in str(folder):
                continue

            return folder / "Fritz-Info.plist"

        # Many default apps are created with application code inside of a
        # folder with the same name as the app name.  If that exists, use
        # that folder.
        if (self.root / self.app_name).exists():
            return self.root / self.app_name / "Fritz-Info.plist"

        return self.root / "Fritz-Info.plist"

    @property
    def fritz_info_path(self) -> Optional[pathlib.Path]:
        """Fritz info path path."""
        return self._fritz_info_path

    @fritz_info_path.setter
    def fritz_info_path(self, value):
        """Fritz info path path."""
        self._fritz_info_path = value

    @property
    def podfile_path(self) -> Optional[pathlib.Path]:
        """Podfile path."""
        return self._podfile_path

    @podfile_path.setter
    def podfile_path(self, value):
        """Podfile path."""
        self._podfile_path = value

    @property
    def xcodeproj_path(self) -> Optional[pathlib.Path]:
        """Path to xcodeproj."""
        return self._xcodeproj_path

    @property
    def fritz_api_key(self) -> Optional[pathlib.Path]:
        """Fritz API Key found in app. """
        if not self.fritz_info_path:
            return None

        return ios_helpers.find_fritz_api_key(self.fritz_info_path)

    @classmethod
    def load(cls, root: pathlib.Path):
        """Search root path and initialize config with found files.

        Args:
            root: Root path of Android project.

        Returns: AndroidConfig.
        """
        config = cls(
            root,
            xcodeproj_path=_find_xcodeproj_path(root),
            app_delegate_path=_find_app_delegate_path(root),
            fritz_info_path=_find_fritz_info_path(root),
            podfile_path=_find_podfile_path(root),
        )

        if not config.xcodeproj_path:
            raise errors.MissingProjectFileError(".xcodeproj")

        if not config.app_delegate_path:
            raise errors.MissingProjectFileError("AppDelegate.swift")

        return config

    def summary(self):
        """Summary of iOS App config."""
        return """
App Name:            {app_name}
Bundle ID:           {bundle_id}
Xcode Project Path:  {xcodeproj_path}
AppDelegate Path:    {app_delegate_path}
Fritz-Info path:     {fritz_info_path}
Podfile path:        {podfile_path}
        """.format(
            app_name=self.app_name,
            bundle_id=self.package_name,
            xcodeproj_path=self.xcodeproj_path,
            app_delegate_path=self.app_delegate_path,
            fritz_info_path=self.fritz_info_path,
            podfile_path=self.podfile_path,
        )



"""
fritz.api_resources.model_version
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
import os
import click
from termcolor import colored

import fritz
from cli import print_utils
from cli import click_utils


# A decorator to pass the model to the current command.
# pylint: disable=invalid-name
pass_model = click_utils.make_pass_decorator(fritz.Model, required=True)
pass_model_optional = click_utils.make_pass_decorator(
    fritz.Model, required=False
)


@click.command()
@click.option("--project-id", help="Restrict results to Project ID")
@click.option(
    "--all-projects", help="List models from all projects", is_flag=True
)
@click.option(
    "--include-archived", help="Include archived models in list", is_flag=True
)
def models(project_id=None, all_projects=False, include_archived=False):
    """List all models."""
    row_format = [
        ("name", "{name:%s}", "Model Name"),
        ("framework", "{framework:%s}", "Framework"),
        ("version", "{version:%s}", "Active Version"),
        ("created", "{created:%s}", "Created"),
        ("updated", "{updated:%s}", "Updated"),
        ("model_id", "{model_id:%s}", "Model ID"),
    ]
    rows = []

    models = fritz.Model.list(project_id=project_id, all_projects=all_projects)
    if not include_archived:
        models = [model for model in models if not model.archived_at]

    if all_projects:
        project_message = "all projects."
    else:
        project = fritz.Project.get(project_id or fritz.project_id)
        project_message = project.name + " project."

    print_utils.notice(
        "\n{num_models} models in {project_message}\n".format(
            num_models=len(models), project_message=project_message
        )
    )

    for model in models:
        row = {
            "name": model.name,
            "framework": model.framework.name,
            "version": model.active_version,
            "created": model.created_date,
            "updated": model.updated_date,
            "model_id": model.id,
        }
        rows.append(row)

    rows.sort(key=lambda x: x["updated"], reverse=True)
    print_utils.formatted_table(row_format, rows)


@click.group(
    cls=click_utils.OptionalArgumentGroup,
    invoke_without_command=True,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.argument("model_id", required=False)
@click.pass_context
def model(ctx, model_id):
    """Commands for working with Fritz Models."""
    model = None
    if model_id:
        model = fritz.Model.get(model_id=model_id)
        ctx.obj = model

    if model and not ctx.invoked_subcommand:
        ctx.invoke(details)
        click.echo(ctx.command.get_help(ctx))
        return

    if not model and not ctx.invoked_subcommand:
        click.echo(ctx.command.get_help(ctx))
        return


@model.command()
@click.option("--description", help="Updated model description")
@click.option("--active-version", type=int, help="Active model version.")
@click.option("--name", help="Name of model")
@click.option(
    "--is-global", help="If true, all projects and accounts can access model."
)
@click.option("--archive", type=bool, help="If true, model will be archived")
@pass_model
def update(model, **updates):
    """Update model.

    Args:
        updates: Dict of updates.
    """
    updates = {
        key: value for key, value in updates.items() if value is not None
    }
    if not updates:
        print_utils.warn("\nNo updates, aborting.\n")
        return

    model = model.update(**updates)
    print_utils.message(model.summary())


@model.command()
@click.option(
    "--version",
    type=int,
    help="Specific version number. If not specified, uses latest",
)
@pass_model
@click.argument("output-dir", required=False)
def download(model, output_dir=None, version=None):
    """Download model version.

    Args:
        model_id: Model ID
        version_number: Optional version to download,
            defaults to active_version.
        output_dir: Optional directory for model download.
    """
    version, downloaded_path = model.download(
        version_number=version, output_dir=output_dir
    )
    if not version:
        print_utils.warn(
            "\nDid not find model version for version" + str(version)
        )

    print_utils.ack(
        (
            "Successfully downloaded {model_name} - {version} "
            "to {downloaded_path}"
        ).format(
            model_name=model.name,
            version=version.version_number,
            downloaded_path=downloaded_path,
        )
    )


@model.command()
@click.option("--deploy", is_flag=True, flag_value=True)
@click.option("--api-key", help="Fritz API Key")
@click.option("--project-id", help="Fritz Project ID")
@click.option(
    "--skip-model-init",
    is_flag=True,
    help="Do not initialize model, just upload bytes.",
)
@click.argument("path")
@pass_model_optional
def upload(
    model,
    path,
    deploy=False,
    api_key=None,
    project_id=None,
    skip_model_init=False,
):

    """Upload model version to Fritz API.

    Args:
        path: Path to model file.
        model_id: Model ID of existing model.
        deploy: If True, will deploy newly uploaded model.
    """
    filename = os.path.basename(path)
    model_file = fritz.frameworks.build_framework_file(
        path,
        # Skipping initalization of the model.  Loading the class (say
        # from Keras) fails a handful of times with custom operations.
        skip_model_cls_init=True,
    )
    model_id = model.id if model else None
    version, snapshot, fritz_model = fritz.ModelVersion.create(
        filename=filename,
        data=model_file.to_bytes(),
        model_id=model_id,
        set_active=deploy,
        api_key=api_key,
        project_id=project_id,
    )

    print_utils.message(fritz_model.summary())
    print_utils.message(version.summary())

    return version, snapshot, model


@model.command()
@click.option("--version-id", help="Existing model version.")
@click.option("--api-key", help="Fritz API Key")
@click.argument("path", required=False)
@click.pass_context
def benchmark(ctx, version_id=None, path=None, api_key=None):
    """Get model grader report for a model version or Keras file.

    Args:
        version_id: Optional ModelVersion ID.
        path: Optional Path to model
    """
    if not version_id and not path:
        print(colored("Version ID or path to model required.", "yellow"))
        return
    if not version_id:
        version, _, _ = ctx.invoke(upload, path=path, skip_model_init=False)
    else:
        version = fritz.ModelVersion.get(version_id, api_key=api_key)

    report = version.benchmark()
    report.summary()


@model.command()
@pass_model
@click.pass_context
def details(ctx, model):
    """Get details about specific model.

    Args:
        id: Model ID to query.
    """
    click.echo("")
    print_utils.message(model.summary())
    click.echo("")
    ctx.forward(versions)


@model.command()
@pass_model
def versions(model):
    """List Model Versions for a specific model_id."""
    versions = fritz.ModelVersion.list(model.id)
    versions_name = print_utils.make_plural(len(versions), "version")
    print_utils.notice(
        "\n{num_versions} model {versions_name} in {model_name}\n".format(
            num_versions=len(versions),
            versions_name=versions_name,
            model_name=model.name,
        )
    )

    row_format = [
        ("filename", "{filename:%s}", "Provided Filename"),
        ("version", "{version:%s}", "Version"),
        ("created", "{created:%s}", "Created"),
        ("model_size", "{model_size:>%s}", "Model Size"),
        ("version_id", "{version_id:%s}", "Version ID"),
        ("loss", "{loss:%s.4}", "Loss"),
    ]
    rows = []

    for version in versions:
        row = {
            "filename": version.provided_filename,
            "version": version.version_number,
            "created": version.created_date,
            "model_size": version.model_size,
            "version_id": version.id,
            "loss": version.loss or "",
        }
        rows.append(row)

    rows.sort(key=lambda x: x["version"], reverse=True)
    print_utils.formatted_table(row_format, rows)


@model.group(invoke_without_command=True)
@click.argument("version_id")
@pass_model
@click.pass_context
def version(ctx, model, version_id):
    """Commands related to specific versions of a model."""
    version = fritz.ModelVersion.get(version_id=version_id)
    ctx.obj = version

    if version and not ctx.invoked_subcommand:
        ctx.invoke(version_details)
        click.echo("")
        click.echo(ctx.command.get_help(ctx))


@version.command("details")
@click.pass_context
def version_details(ctx):
    """Get Details for a specific Model Version.

    Args:
        version_id: Version ID to query.
    """
    print_utils.message(ctx.obj.summary())



"""
cli.app.apps
~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
import os

import click

import fritz

from cli.app import app_setup
from cli import print_utils
from cli import click_utils

# pylint: disable=invalid-name
pass_app = click_utils.make_pass_decorator(fritz.App)


@click.command()
@click.option("--project-id", help="Restrict results to specific project.")
@click.option("--apk-id", help="Restrict results to APK ID or Bundle ID.")
@click.option("--platform", help="Restrict results to specific platform.")
@click.option(
    "--all-projects", is_flag=True, help="Return results from all projects"
)
def apps(project_id=None, apk_id=None, platform=None, all_projects=False):
    """List all apps."""
    click.echo("")
    apps = fritz.App.list(
        project_id=project_id,
        apk_id=apk_id,
        platform=platform,
        all_projects=all_projects,
    )
    if all_projects:
        project_message = "all projects."
    else:
        project = fritz.Project.get(project_id or fritz.project_id)
        project_message = "{project_name} project.".format(
            project_name=project.name
        )

    apps_name = print_utils.make_plural(len(apps), "app")
    print_utils.notice(
        "{num_apps} {apps_name} in {project_message}\n".format(
            num_apps=len(apps),
            apps_name=apps_name,
            project_message=project_message,
        )
    )

    row_format = [
        ("name", "{name:%s}", "App Name"),
        ("apk_id", "{apk_id:%s}", "APK ID"),
        ("platform", "{platform:%s}", "Platform"),
        ("created", "{created:%s}", "Created"),
        ("project_name", "{project_name:%s}", "Project Name"),
        ("api_key", "{api_key:%s}", "API Key"),
        ("app_id", "{app_id:%s}", "App ID"),
    ]
    rows = []
    projects = {}
    for app in apps:
        project = projects.setdefault(
            app.project_id, fritz.Project.get(project_id=app.project_id)
        )
        row = {
            "name": app.app_name,
            "apk_id": app.apk_id,
            "platform": app.platform,
            "created": app.created_date,
            "project_name": project.name,
            "api_key": app.api_key,
            "app_id": app.id,
        }
        rows.append(row)

    rows.sort(key=lambda x: x["created"], reverse=True)
    print_utils.formatted_table(row_format, rows)


@click.group(
    cls=click_utils.OptionalArgumentGroup,
    invoke_without_command=True,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.argument("app-id", required=False)
@click.pass_context
def app(ctx, app_id=None):
    """Commands for working with Fritz Apps.

    Args:
        app_id: App ID to query.
    """

    if app_id:
        ctx.obj = fritz.App.get(app_id=app_id)

    if ctx.obj and not ctx.invoked_subcommand:
        ctx.invoke(details)
        click.echo("")
        click.echo(ctx.command.get_help(ctx))
        return

    if not ctx.obj and not ctx.invoked_subcommand:
        click.echo(ctx.command.get_help(ctx))
        return


app.add_command(app_setup.setup)


@app.command()
@pass_app
def details(app):
    """Get details of a specific app."""
    print_utils.message(app.summary())


@app.group()
def pod():
    """Commands for working with Fritz cocoapods."""


@pod.command()
@click.option("--clean", is_flag=True)
def install(clean):
    """Install the Fritz Cocoapod.

    Args:
        clean (bool): if present, delete cached cocoapods and update the
            repo as well.
    """
    commands = []
    if clean:
        commands.extend(
            [
                "pod repo update --verbose",
                "pod clean --verbose",
                "pod cache clean --all --verbose",
            ]
        )
    commands.append("pod install --verbose")
    for command in commands:
        print(command)
        os.system(command)



"""
cli.project
~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""

import click
import fritz
from cli import print_utils
from cli import fritz_object_selector
from cli.config import update_config
from cli import click_utils

# pylint: disable=invalid-name
pass_project = click_utils.make_pass_decorator(fritz.Project)


@click.command()
def projects():
    """List all Projects for current account."""
    row_format = [
        ("name", "{name:%s}", "Project Name"),
        ("created", "{created:%s}", "Created"),
        ("project_id", "{project_id:%s}", "Project ID"),
    ]
    rows = []

    projects = fritz.Project.list()
    projects_name = print_utils.make_plural(len(projects), "project")
    print_utils.notice(
        "\n {num_projects} {projects_name}\n".format(
            num_projects=len(projects), projects_name=projects_name
        )
    )

    for project in projects:
        row = {
            "name": project.name,
            "created": project.created_date,
            "project_id": project.id,
        }
        rows.append(row)

    rows.sort(key=lambda x: x["created"], reverse=True)
    print_utils.formatted_table(row_format, rows)


@click.group(
    cls=click_utils.OptionalArgumentGroup,
    invoke_without_command=True,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.argument("project-id", required=False)
@click.pass_context
def project(ctx, project_id=None):
    """Commands for working with Fritz Projects."""

    if project_id:
        ctx.obj = fritz.Project.get(project_id=project_id)

    if ctx.obj and not ctx.invoked_subcommand:
        ctx.invoke(details)
        click.echo("")
        click.echo(ctx.command.get_help(ctx))
        return

    if not ctx.obj and not ctx.invoked_subcommand:
        click.echo(ctx.command.get_help(ctx))
        return


@project.command()
@pass_project
def details(project):
    """Get details of project.

    Args:
        id: Optional Project ID. If not specified, will use project ID
            specified in Fritz Config.
    """
    project.summary()


@project.command()
@click.argument("project_id", required=False)
def set_active(project_id):
    """Set Project ID as the active project in fritz config."""
    if project_id:
        update_config(project_id=project_id)
        return

    projects = fritz.Project.list()
    projects_name = print_utils.make_plural(len(projects), "project")
    print_utils.notice(
        "\n {num_projects} {projects_name}\n".format(
            num_projects=len(projects), projects_name=projects_name
        )
    )
    project = fritz_object_selector.choose_project(projects)
    update_config(project_id=project.id)



"""
fritz.api_resources.project
~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""

import fritz
from fritz.api_resources import fritz_object


class Project(fritz_object.FritzObject):
    """Fritz Project object."""

    OBJECT_NAME = "Project"

    def summary(self):
        """Print summary of object."""
        details = [("Created At", self.created_date)]
        return self._summary(self.name, details)

    @classmethod
    def get(cls, project_id=None, api_key=None):
        """Get project by id

        Args:
            project_id (str): ID of proejct
            api_key (str): Optional API Key to use.

        Returns: Project
        """
        project_id = project_id or fritz.project_id
        url = "/project/{project_id}".format(project_id=project_id)
        response = cls.client_get(url, api_key=api_key)
        return fritz.utils.convert_to_fritz_object(response)

    @classmethod
    def list(cls, api_key=None):
        """Get project by id

        Args:
            api_key (str): Optional API Key to use.

        Returns: Project
        """
        url = "/project"
        response = cls.client_get(url, api_key=api_key)

        return [
            fritz.utils.convert_to_fritz_object(project)
            for project in response["projects"]
        ]



"""
fritz.fritz_object
~~~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
from typing import Dict
import json
import fritz
from fritz.api_client import FritzClient

# Keys that are accessed with "uid" in data.
UID_KEYS = ["app_id", "model_id", "project_id", "snapshot_id", "version_id"]


class FritzObject(object):
    """An object from the API.

    Subclassed objects are used to define interactions with the API.
    """

    def __init__(self, **kwargs):
        self._data = kwargs

    def __repr__(self):
        identity_parts = [type(self).__name__]

        if isinstance(self.uid, str):
            identity_parts.append("uid=%s" % (self.uid))

        unicode_repr = "<%s> JSON: %s" % (" ".join(identity_parts), str(self))

        return unicode_repr

    def __str__(self):
        return json.dumps(self._data, sort_keys=True, indent=2)

    def __getattr__(self, key):
        if key[0] == "_":
            raise AttributeError(key)

        # We refer to ids externally as "project_id", "model_id", etc, but
        # all underlying data is stored as "project_uid", "model_uid", etc.
        # Enforcing accessing this with id rather than uid.
        split_key = key.split("_")
        if len(split_key) == 2 and split_key[1] == "uid":
            raise AttributeError(key)

        if key in UID_KEYS:
            key = "_".join([split_key[0], "uid"])

        try:
            return self._data[key]
        except KeyError as err:
            raise AttributeError(*err.args)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.__dict__ == other.__dict__

    @property
    def data(self) -> Dict:
        """Underlying data backing object."""
        return self._data

    @property
    def id(self):
        """ID of object.

        Returns: str
        """
        # Note: Internally, IDs are stored as UIDs.  However, externally
        # these are consistently referred to as IDs (i.e. model_id).
        return self.uid

    @property
    def created_date(self):
        """Date when object was created."""
        date, time = self.created_at.split("T")
        return date + " " + time[:5]

    @property
    def updated_date(self):
        """Date when object was created."""
        date, time = self.updated_at.split("T")
        return date + " " + time[:5]

    def _summary(self, object_name, details):
        header_fmt = "{key:>%s}: {value}" % len(self.OBJECT_NAME)
        title = header_fmt.format(key=self.OBJECT_NAME, value=object_name)
        id_row = header_fmt.format(key="ID", value=self.id)
        header_line = "-" * max(len(title), len(id_row))
        lines = [title, id_row, header_line]
        max_name = max(len(name) for name, _ in details)
        for name, value in details:
            row_fmt = "  {name:%s}: {value}" % (max_name + 2)
            lines.append(row_fmt.format(name=name, value=value))

        return "\n".join(lines)

    @staticmethod
    def client_post(url, api_key=None, params=None):
        """Base method to perform POST requests.

        Does not attempt to convert response into object. Implementing classes
        should handle that.

        Args:
            url (str): URL to query.
            api_key (str): Optional Fritz API key.
            params (dict): JSON body to pass to request.

        Returns: dict response if successful.
        """
        if not fritz.api_key and not api_key:
            raise fritz.errors.FritzNotInitializedError()

        client = FritzClient(api_key or fritz.api_key)
        response = client.post(url, json=params)

        return response

    @staticmethod
    def client_put(url, api_key=None, params=None):
        """Base method to perform PUT requests.

        Does not attempt to convert response into object. Implementing classes
        should handle that.

        Args:
            url (str): URL to query.
            api_key (str): Optional Fritz API key.
            params (dict): JSON body to pass to request.

        Returns: dict response if successful.
        """
        if not fritz.api_key and not api_key:
            raise fritz.errors.FritzNotInitializedError()

        client = FritzClient(api_key or fritz.api_key)
        response = client.put(url, json=params)

        return response

    @staticmethod
    def client_get(url, api_key=None, params=None):
        """Base method to perform GET requests.

        Does not attempt to convert response into object. Implementing classes
        should handle that.

        Args:
            url (str): URL to query.
            api_key (str): Optional Fritz API key.
            params (dict): Query parameters to pass to request.

        Returns: dict response if successful.
        """
        if not fritz.api_key and not api_key:
            raise fritz.errors.FritzNotInitializedError()

        client = FritzClient(api_key or fritz.api_key)

        response = client.get(url, params=params)

        return response



"""
fritz.errors
~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
# Right now, I'm going to tie in our normal exceptions with click.
# this shouldn't affect too much? I don't love it, but the priority for
# me is showing readable errors.
import click


class FritzError(click.ClickException):
    """Generic Exception class for Fritz Errors."""

    def __init__(self, message, status_code=None):
        """
        Args:
            message (str): Error message
            status_code (Optional[int]): Status Code of response.
        """
        super().__init__(message)
        self._message = message
        self.status_code = status_code

    def __str__(self):
        return self._message or "<empty message>"

    def __repr__(self):
        cls_name = self.__class__.__name__

        return "{cls_name}(message={msg} status_code={status_code})".format(
            cls_name=cls_name, msg=self._message, status_code=self.status_code
        )


class FritzNotInitializedError(FritzError):
    """Error when Python SDK not initialized."""

    def __init__(self):
        message = (
            "`fritz.configure` not called.  Please call "
            "`fritz.configure` with your API Key and Project ID."
        )
        super().__init__(message)


class MissingFritzConfigError(FritzError):
    """Missing Fritz configuration file."""

    def __init__(self, path):
        message = """

No Fritz Configuration file exists at {path}

Please create a file at {path} with the follinwg contents.
[default]
api_key = <api_key>
project_id = <project_id>

This is easy with the Fritz CLI:

$ fritz config --api-key <api_key> --project-id <project_id>
        """
        super().__init__(message.format(path=path))


class InvalidFritzConfigError(FritzError):
    """Error when Fritz config contains invalid options."""

    def __init__(self, path):
        message = """
        Fritz configuration file at {path} contains invalid options

        Please check your Fritz configuration file at ~/.fritz.

        To update the config, run:

        $ fritz config udpate
        """
        super().__init__(message.format(path=path))


class ConnectionError(FritzError):
    """Could not not connect to server."""

    def __init__(self, path):
        message = "Error connecting to path {path}. Please check your connection and try again."
        super().__init__(message.format(path=path))


class ArgumentRequiredError(FritzError):
    """Argument required"""

    def __init__(self, id_name):
        message = "{id_name} argument required."
        super().__init__(message.format(id_name=id_name))


class MissingProjectFileError(FritzError):
    """Could not find expected file."""

    def __init__(self, file_name):
        message = """
        Project missing {file_name}.

        Check to make sure project root is pointing to directory containing
        your mobile app code.
        """.format(
            file_name=file_name
        )

        super().__init__(message)



"""Shared methods for app setup."""

import click
import fritz
from cli import print_utils


def pause_confirmation():
    """Pause flow until the user finishes the previous step"""

    click.confirm("\nContinue? Press any key", default="y", show_default=False)


def setup_app_info(app_name, package_name, platform):
    """Create the app automatically

    Args:
        app_name
        package_name
        platform

    Returns:
        the newly created app from the api.
    """

    app_name = click.prompt(
        "Please enter the name of your app",
        default=app_name,
        show_default=True,
    )
    while not app_name:
        app_name = click.prompt(
            "Please enter the name of your app",
            default=app_name,
            show_default=True,
        )

    package_name = click.prompt(
        "Please enter the package name of your app",
        default=package_name,
        show_default=True,
    )
    while not package_name:
        package_name = click.prompt(
            "Please enter the package name of your app",
            default=package_name,
            show_default=True,
        )
    response = fritz.App.create(
        api_key=fritz.api_key,
        app_name=app_name,
        apk_id=package_name,
        platform=platform,
    )

    print_utils.ack(
        "\n{app_name} ({package_name}) created successfully!".format(
            app_name=app_name, package_name=package_name
        )
    )

    return response



"""
Utilities to help running click commands.

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""

import click
from click import decorators


def make_pass_decorator(object_type, required=True):
    """Finds object of type object_type in ctx and passes as first argument.

    Args:
        object_type: The type of the object to pass.
        required: If set to true, object must be present or an error is thrown.

    Returns: Decorator to find function.
    """

    def decorator(f):
        def new_func(*args, **kwargs):
            ctx = decorators.get_current_context()
            obj = ctx.find_object(object_type)
            if obj is None and required:
                click.echo(ctx.command.get_help(ctx))
                click.echo("")
                raise click.ClickException(
                    "Managed to invoke callback without a "
                    "context object of type %r existing" % object_type.__name__
                )
            return ctx.invoke(f, obj, *args, **kwargs)

        return decorators.update_wrapper(new_func, f)

    return decorator


class OptionalArgumentGroup(click.Group):
    """Group that supports subcommands with optional parent arguments.

    For instance, for the fritz model <model_id> upload <file> command
    the <model_id> command is optional.  Without this group, "upload"
    is treated as the model ID.
    """

    def parse_args(self, ctx, args):
        """Parse args."""
        if not args:
            super().parse_args(ctx, args)
            return

        if args[0] in self.commands:
            if len(args) == 1 or args[1] not in self.commands:
                args.insert(0, "")
        super().parse_args(ctx, args)



"""
cli.fritz_object_selector

Functions to select a single fritz object from a click prompt.
"""
from typing import List, Optional
import click
import fritz
from cli import print_utils


def _row_chooser_prompt(rows, row_format, include_none=False):

    row_format = [("index", "{index:%s}", "Index")] + row_format

    new_objects = []

    # Giving option to not choose an item in the list
    if include_none:
        row = {key: "----" for key, _, _ in row_format}
        row["index"] = "[0]"
        # Setting the field of the first object as the value None
        row[row_format[1][0]] = "None"
        new_objects.append(row)

    for i, obj in enumerate(rows):
        object_with_index = dict(**obj)
        object_with_index["index"] = "[" + str(i + 1) + "]"
        new_objects.append(object_with_index)

    print_utils.formatted_table(row_format, new_objects)
    options = click.IntRange(
        min=0 if include_none else 1, max=len(new_objects)
    )
    index = click.prompt("Select an index", type=options)
    if not index:
        return None

    return index - 1


def choose_app(
    apps: List[fritz.App], include_none: bool = False
) -> Optional[fritz.App]:
    """Prints app table to choose a single app from.

    Args:
        apps: Apps
        include_none: If True, includes option to not select app.

    Returns: Selected app
    """
    row_format = [
        ("name", "{name:%s}", "App Name"),
        ("apk_id", "{apk_id:%s}", "APK ID"),
        ("created", "{created:%s}", "Created"),
        ("project", "{project:%s}", "Project"),
    ]
    apps.sort(key=lambda x: x.created_at, reverse=True)
    projects = {}
    rows = []
    for app in apps:
        if app.project_id not in projects:
            projects[app.project_id] = fritz.Project.get(
                project_id=app.project_id
            )
        row = {
            "name": app.app_name,
            "apk_id": app.apk_id,
            "platform": app.platform,
            "created": app.created_date,
            "project": projects[app.project_id].name,
        }
        rows.append(row)

    index = _row_chooser_prompt(rows, row_format, include_none=include_none)

    if index is None:
        return None

    return apps[index]


def choose_model(
    models: List[fritz.Model], include_none: bool = False
) -> Optional[fritz.Model]:
    """Prints model table to choose a single model from.

    Args:
        models: Models
        include_none: If True, includes option to not select model.

    Returns: Selected model
    """
    row_format = [
        ("name", "{name:%s}", "Name"),
        ("model_format", "{model_format:%s}", "Format"),
        ("created", "{created:%s}", "Created"),
        ("updated", "{updated:%s}", "Updated"),
        ("active_version", "{active_version:%s}", "Active Version"),
    ]
    models.sort(key=lambda x: x.updated_at, reverse=True)
    rows = []
    for model in models:
        row = {
            "name": model.name,
            "model_format": model.model_format,
            "created": model.created_date,
            "updated": model.updated_date,
            "active_version": model.active_version,
        }
        rows.append(row)

    index = _row_chooser_prompt(rows, row_format, include_none=include_none)

    if index is None:
        return None

    return models[index]


def choose_project(
    projects: List[fritz.Project], include_none: bool = False
) -> Optional[fritz.Project]:
    """Prints project table and prompts choice of one.

    Args:
        projects: Projects
        include_none: If True, includes option to not select project.

    Returns: Selected project
    """
    row_format = [
        ("name", "{name:%s}", "Project Name"),
        ("created", "{created:%s}", "Created"),
        ("project_id", "{project_id:%s}", "Project ID"),
    ]
    rows = []

    projects.sort(key=lambda x: x.created_at, reverse=True)

    for project in projects:
        row = {
            "name": project.name,
            "created": project.created_date,
            "project_id": project.id,
        }
        rows.append(row)

    index = _row_chooser_prompt(rows, row_format, include_none=include_none)
    if index is None:
        return None

    return projects[index]



# pylint: disable=missing-docstring
import sys
import os
from os import path

from setuptools import setup
from setuptools import find_packages
from setuptools.command.test import test as TestCommand


def get_readme_contents():
    # Setup with the description in the README file
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
        long_description = f.read()
        return long_description

    return None


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ["--boxed"]

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        if isinstance(self.pytest_args, str):
            self.pytest_args = self.pytest_args.split(" ")
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


here = os.path.abspath(os.path.dirname(__file__))
os.chdir(here)


tests_require = [
    "astroid >= 2.1.0",
    "pytest >= 4.0.0",
    "pytest-runner",
    "pytest-pylint",
    "pytest-xdist",
    "pytest-cov",
    "pytest-mock",
    "requests-mock",
    "pylint",
    "requests_toolbelt",
    "Sphinx",
    "keras",
    "tensorflow <= 1.13.1",
    "coremltools",
    "numpy < 1.17",
]


setup(
    name="fritz",
    version="2.1.5",
    description="Fritz Machine Learning Library.",
    long_description=get_readme_contents(),
    long_description_content_type="text/markdown",
    url="https://github.com/fritzlabs/fritz-python",
    project_urls={
        "Documentation": "https://docs.fritz.ai/cli/index.html",
        "Source Code": "https://github.com/fritzlabs/fritz-python",
        "Bug Tracker": "https://github.com/fritzlabs/fritz-python/issues",
    },
    keywords="machine learning, training, app development",
    author="Fritz Engineering",
    author_email="engineering@fritz.ai",
    license="MIT",
    zip_safe=False,
    packages=find_packages(),
    install_requires=[
        "requests",
        "click",
        "click-plugins",
        "termcolor",
        "pbxproj",
    ],
    extras_require={"train": ["keras", "tensorflow"]},
    tests_require=tests_require,
    cmdclass={"test": PyTest},
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points={"console_scripts": ["fritz = cli:main"]},
)



"""
cli.config
~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
import configparser
import os
from pathlib import Path
import click

from fritz import configuration
import fritz


_HOME = str(Path.home())
CONFIG_PATH = os.path.join(_HOME, ".fritz")


def update_config(path: str = CONFIG_PATH, **updates):
    """Update Fritz configuration file, creating if it does not exist.

    Args:
        path (str): Path of config file.
        updates (dict): Optional keys to update specified by CLI options.

    Return: ConfigParser
    """
    fritz_config = configuration.load_config_file(path=path)

    if updates:
        try:
            defaults = configuration.get_credentials(fritz_config)
        except configparser.NoSectionError:
            defaults = {}

        changes = []
        for key in ["api_key", "project_id"]:
            if key not in updates:
                continue

            original = defaults.get(key)
            new = updates[key]
            changes.append(
                "{key}: {original} -> {new}".format(
                    key=key, original=original, new=new
                )
            )

        print(
            "\n".join(["Updating config with provided variables.", *changes])
        )

        fritz_config = configuration.update_credentials(
            fritz_config, **updates
        )
        configuration.write_config_file(fritz_config, path=path)
        return fritz_config

    click.echo(
        """
    Configure the Fritz CLI.
    ------------------------

    This command will add a ~/.fritz configuration file with your API Key
    and default Project ID.

    The API Key is used to authenticate with the Fritz servers, and the
    Project ID is used to determine whcih project to use when creating and
    benchmarking new models.

    To find your API Key and Project ID, log in to Fritz (https://app.fritz.ai)
    and go to the "Training" tab in you project. Enter the listed API Key and
    Project ID in the following prompts:

    """
    )
    api_key = click.prompt("Fritz API Key", type=str)
    project_id = click.prompt("Fritz Project ID", type=str)

    updates = {"api_key": api_key, "project_id": project_id}
    configuration.update_credentials(fritz_config, **updates)
    configuration.write_config_file(fritz_config, path=path)
    return fritz_config


@click.group(invoke_without_command=True)
@click.pass_context
def config(ctx):
    """Fritz configuration settings.

    Args:
        update: Optional flag to update config file.
    """
    if ctx.invoked_subcommand is not None:
        # Don't want to run default flow if specifically trying to update.
        return

    try:
        configuration.init_config()
    except fritz.errors.InvalidFritzConfigError:
        update_config()

    message = "Fritz Configuration"
    print(message)
    print("=" * len(message))
    print("API Key: {api_key}".format(api_key=fritz.api_key))
    print("Project ID: {project_id}".format(project_id=fritz.project_id))


@config.command()
@click.option("--api-key", help="Fritz API Key", default=None)
@click.option("--project-id", help="Fritz Project ID", default=None)
def update(api_key, project_id):
    """Fritz configuration settings.

    Args:
        update: Optional flag to update config file.
    """
    updates = {"api_key": api_key, "project_id": project_id}
    updates = {key: value for key, value in updates.items() if value}
    update_config(**updates)
    fritz.configure()

    message = "Fritz Configuration"
    print(message)
    print("=" * len(message))
    print("API Key: {api_key}".format(api_key=fritz.api_key))
    print("Project ID: {project_id}".format(project_id=fritz.project_id))



"""
fritz.api_client
~~~~~~~~~~~~~~~~
This module contains the FritzClient for interacting with the Fritz API.
"""
import requests

import fritz
import fritz.errors


class FritzClient(object):
    """Fritz Client to interact with Fritz API."""

    def __init__(self, api_key=None, api_base=None):
        """
        Args:
            api_key (str): Account API Key for Fritz API.
            base_url (str): URL base of Fritz API.
        """
        self._api_key = api_key or fritz.api_key
        self._api_base = api_base or fritz.api_base

    def put(
        self, url, params=None, data=None, json=None, files=None, headers=None
    ):
        """PUT Request to Fritz API.

    Args:
            url (str): URL to request.
            params (dict): Query parameters
            data (dict): Content
            files (dict): Files to upload
            headers (dict): Headers to add to request.
            json (dict): the json content

        Returns: Dict of response data.
        """
        path = self._api_base + url
        base_headers = {"Authorization": self._api_key}
        base_headers.update(headers or {})
        response = requests.put(
            path,
            params=params,
            data=data,
            json=json,
            files=files,
            headers=base_headers,
        )
        try:
            response.raise_for_status()
        except requests.exceptions.ConnectionError:
            raise fritz.errors.ConnectionError(path)
        except requests.exceptions.RequestException:
            error = response.json()
            raise fritz.errors.FritzError(
                error.get("message"), status_code=response.status_code
            )

        return response.json()

    # Note: Chris - these feel a bit duplicated. Wonder if we can combine this
    # in some way.
    def post(
        self, url, params=None, data=None, json=None, files=None, headers=None
    ):
        """POST Request to Fritz API.

    Args:
            url (str): URL to request.
            params (dict): Query parameters
            data (dict): Content
            files (dict): Files to upload
            headers (dict): Headers to add to request.
            json (dict): the json content

        Returns: Dict of response data.
        """
        path = self._api_base + url
        base_headers = {"Authorization": self._api_key}
        base_headers.update(headers or {})

        response = requests.post(
            path,
            params=params,
            data=data,
            json=json,
            files=files,
            headers=base_headers,
        )
        try:
            response.raise_for_status()
        except requests.exceptions.ConnectionError:
            raise fritz.errors.ConnectionError(path)
        except requests.exceptions.RequestException:
            error = response.json()
            raise fritz.errors.FritzError(
                error.get("message"), status_code=response.status_code
            )

        return response.json()

    def get(self, url, params=None, headers=None):
        """GET Request to Fritz API.

        Args:
            url (str): URL to request.
            params (dict): Query parameters
            headers (dict): Headers to add to request.

        Returns: Dict of response data.
        """
        path = self._api_base + url
        base_headers = {"Authorization": self._api_key}
        base_headers.update(headers or {})
        try:
            response = requests.get(path, params=params, headers=base_headers)
            response.raise_for_status()
        except requests.exceptions.ConnectionError:
            raise fritz.errors.ConnectionError(path)
        except requests.exceptions.RequestException:
            error = response.json()
            raise fritz.errors.FritzError(
                error.get("message"), status_code=response.status_code
            )

        return response.json()



"""
Fritz CLI

usage:
   $ fritz model upload my_keras_file.h5
   $ Model Details
     =============
     Model ID:       6bd0fbb59f264a3c8b79e963558848e2
     Model Name:      multi_person_mobilenet_v1_075_float
     Active version:  1

     Model Version Details
     ======================
     Model Version ID:  420fd0b3ef5e4fc48a5b535244d0c0a3
     Version Number:     1
   $ fritz model benchmark --version-id 420fd0b3ef5e4fc48a5b535244d0c0a3
     ------------------------
     Fritz Model Grade Report
     ------------------------
     Core ML Compatible:              True
     Predicted Runtime (iPhone X):    31.4 ms (31.9 fps)
     Total MFLOPS:                    686.90
     Total Parameters:                1,258,580

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""

from pkg_resources import iter_entry_points
import click
from click_plugins import with_plugins
import fritz
from cli.config import config
from cli.model import model
from cli.model import models
from cli.app import apps
from cli import project as project_cli


class FritzNotConfiguredError(click.UsageError):
    """Error raised when Fritz CLI has not been configured."""

    def __init__(self, ctx=None):
        message = (
            "Please configure Fritz API Key and Project ID:\n"
            "\n  $ fritz config\n"
        )
        super().__init__(message, ctx=ctx)


@with_plugins(iter_entry_points("fritz.plugins"))
@click.group()
@click.option("--skip-fritz-configure", default=False, is_flag=True)
@click.pass_context
def main(ctx, skip_fritz_configure=False):
    """Fritz CLI."""
    # fritz config subcommand will enable fritz.configure() to work.
    # If that is the subcommand, don't call fritz.configure().
    if ctx.invoked_subcommand == "config":
        return
    if skip_fritz_configure:
        return

    fritz.configure()


main.add_command(config)
main.add_command(models)
main.add_command(model)
main.add_command(apps.apps)
main.add_command(apps.app)
main.add_command(project_cli.project)
main.add_command(project_cli.projects)
